/*------------------------------------------------------------------------
			CLIENT FOR THE CERN PHONEBOOK

Module:		ph.c
Purpose:	Main Program
Author:		Josef Pellizzari / CN
Last Changed:
Change record:  11.Dec.1999 Andreas Wagner IT/US   


************************** DISCLAIMER FROM THE AUTHOR ********************
This program has been copied together from modules I used somewhere
else. It has been handed over to Bernd Pollermann ("The Phone People")
so that he doesn't have to write it from scratch. It will probably not
compile on all platforms and its maintenance is the responsibility
of the phone people - certainly not mine! So - please complain there!
**************************************************************************

Description:
This client opens a socket and then tries to connect to the phone server.
If this operation is succesfull it concatenates the username and all command
line arguments and transmits this as the request to the server. Afterwards
it reads in a loop the result from the server until the socket is closed
by the server.

The server and the port number of the socket are hardcoded but can be
overridden from the command line with the '-server <hostname>' and
'-port <Port-Number>' options. As soon as the client finds the first option
it does not expect, i.e. anything besides the options mentioned directly above,
it assumes this as part of the beginning of the request and concats the
remaining argument without any further checks.

Changes:
10.12.96	J. Pellizzari / CERN
		Since the server crashes to often we introduce a spare
		server to contact if the initial one fails. To get rid
		of that for all times we introduce an array of servers
		which might be queried.

14.Jan.1999     Andreas Wagner  CERN IT/US 
                - debugged the -port and the -server option: When the
                  port number is specified no looping over array of 
                  servers
                - if servername is specified also the port has to be
                  specified as well
                - changed default servername from 'wwwcn1' to 'consult'

20.Jan.1999     Andreas Wagner CERN IT/US
                - default server 'sunldap'.
                - if all 3 phone-servers on 'sunldap are down, then using 
                  the 3 servers on 'consult'
                - option -verbose to get detailed output  

------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <unistd.h>
#include <pwd.h>
#include <signal.h>

#include "ph-comm.h"

/*------------------------------------------------------------------------
Defaults for server hostname and port number.
ATTENTION:	Both by the way the command overrides this entries
		and the program logic it's strictly required that
		there is at least one default entry present which has
		to be followed by the NULL/0 line.
------------------------------------------------------------------------*/
struct	{
		char*	hostname;
		int	port;
	}
	servers[] = 
	{
		{	"phonea.cern.ch",	6664	},
		{	"phonea.cern.ch",	6665	},
		{	"phonea.cern.ch",	6666	},
		{	NULL,			0	},
	};


struct	{
		char*	hostname;
		int	port;
	}
	backup_servers[] = 
	{
		{	"phoneb.cern.ch",	6664	},
		{	"phoneb.cern.ch",	6665	},
		{	"phoneb.cern.ch",	6666	},
		{	NULL,			0	},
	};



/*------------------------------------------------------------------------
The size of the buffer, i.e. what we can get out/swallow with a
single read/write call.
------------------------------------------------------------------------*/
#define	MAX_MSG		4096

/*------------------------------------------------------------------------
Prototypes
------------------------------------------------------------------------*/
static	void usage ( void );
static  void CatchSignal(int sgn);
static	void SetupSignals();

/*------------------------------------------------------------------------
Global, may be closed upon signal.
------------------------------------------------------------------------*/
int	sock = -1;

int test = 0;  /* */
int userdefined_port   = 0; /* */
int userdefined_server = 0; /* */

int verbose = 0;
int this_serverindex;

int return_code=0;  /* return code is zer as default */

/*------------------------------------------------------------------------
The whole program
------------------------------------------------------------------------*/
int	main(int argc, char **argv)
{
	char	msg[MAX_MSG+1];
	int	i, j, k, l,
		option_known,
		first_server,
		n_servers;

        char* error_string="Error: ";

	for (n_servers = 0; servers[n_servers].hostname; n_servers++);

	/*----------------------------------------------------------------
	Make sure that there are any command line arguments
	***
	if (argc <= 1)	{ usage(); exit(1); };
	----------------------------------------------------------------*/

	SetupSignals();

	/*----------------------------------------------------------------
	Find out if there are valid server/port options and where the
	query to the phone server starts
	----------------------------------------------------------------*/


	for (i = 1, option_known=1; i<argc;i++ )
	{       
		if (!strcmp(argv[i], "-verbose"))
		{
                  verbose=1;
                  argv[i] = " "; 
		};
		if (!strcmp(argv[i], "-port"))
		{
			if (((i+1) >= argc) || (servers[0].port=atoi(argv[i+1])) <= 0)
			{
				usage(); exit(1);
			}
                        this_serverindex=0;
                        userdefined_port = 1;
			servers[1].hostname = NULL;
			i += 1;
                        argv[i-1]=" ";
                        argv[i]=" ";
		}
		if (!strcmp(argv[i], "-server"))
		{
			if (((i+1) >= argc) ||
			    (servers[0].hostname=argv[i+1]) == NULL ||
			    servers[0].hostname[0] == '\0')
			{
				usage(); exit(1);
			}
			servers[1].hostname = NULL;
                        userdefined_server = 1;
			i += 1;
                        argv[i-1]=" ";
                        argv[i]=" ";
		}
	}
 
        /*----------------------------------------------------------------
          Check that if server is specified also the port number has to be     
        ----------------------------------------------------------------*/
         

        if (userdefined_server == 1 && userdefined_port == 0) 
        {                     
            fprintf(stderr,"The port number is also required when you specify a servername !\n");
            exit(1);
	}


	/*----------------------------------------------------------------
	Make sure that there are any command line arguments after the
	server/port options.

	if ((argc - 2*userdefined_server - 2*userdefined_port ) <= 1) { usage(); exit(1); };

	----------------------------------------------------------------*/


	/*----------------------------------------------------------------
	Connect to the server.
	----------------------------------------------------------------*/
 
        /* don't loop over servers if user specified port*/     
        if (userdefined_port==1) {

                 if (verbose) 
                 { 
                     printf("---------------------------------------");
                     printf("-----------------------------------------\n"); 
                     printf(" Using port number %d",servers[0].port); 

                     if (userdefined_server==1) 
                     { 
                        printf(" and server %s",servers[0].hostname);
                     };

                     printf("\n---------------------------------------");
                     printf("-----------------------------------------\n"); 
		 }

                 sock = connect_port(servers[0].hostname,servers[0].port);

                 if (sock < 0)
                 {
                      fprintf(stderr, "Connect Error: server %s at port %d\n",
                      servers[0].hostname,
                      servers[0].port);
                      exit(1);
		 }
	} else {

            srand(time(NULL)*getpid());
            first_server = rand() % n_servers;
   

         for (k=1; k<3 && sock < 0; k++)   
         {
          for (j=0, sock=-1; j<n_servers && sock < 0; j++)
            {
	        if (verbose) 
                 { 
                     printf("---------------------------------------");
                     printf("-----------------------------------------\n"); 
                   printf("Connecting to %s:%d ... ", 
                   servers[(first_server+j)%n_servers].hostname,
                   servers[(first_server+j)%n_servers].port);
                     printf("\n---------------------------------------");
                     printf("-----------------------------------------\n"); 
                   fflush(stdout);        
		 }

                 this_serverindex=(first_server+j)%n_servers;
                 sock = connect_port(
                 servers[(first_server+j)%n_servers].hostname,
                 servers[(first_server+j)%n_servers].port);
                 /*
                 printf("%s\n", (sock < 0) ? "failed" : "done");
                 */
	         if (verbose && sock < 0) 
                 {
                     fprintf(stderr, "Connect Error: server %s at port %d\n",
                     servers[this_serverindex].hostname,
                     servers[this_serverindex].port);
                 }
            } 
           /* testing if sucessfully connected otherwise using backup_servers */
           if (sock < 0) 
           {  
              if (verbose) 
              { 
                     printf("---------------------------------------");
                     printf("-----------------------------------------\n"); 
                 fprintf(stdout,"Connect Warning: cannot connect to %s, trying %s\n", 
                                 servers[1].hostname, backup_servers[1].hostname);
	      } ;  

              for (l= 0; backup_servers[l].hostname;l++) 
              {
                servers[l].hostname=backup_servers[l].hostname;
                servers[l].port=backup_servers[l].port;
              } ; 
           } ;
         } ; /* end for(k=1;... */ 
	
            if (sock < 0)
            {
              fprintf(stderr, "Connect Error: Xwho servers seem to be down. Please try later.\n");

              /*
              for (j=0; servers[j].hostname; j++)
              fprintf(stderr, "        %s:%d\n",
              servers[j].hostname,
              servers[j].port);
              */

              exit(1);
            } 
        }
   
	/* Request format: "UID Keyword(s)"	*/
	/*----------------------------------------------------------------
	The server wants to know the user idendity and then the query.
	Here we start to build the request.
	----------------------------------------------------------------*/
	{
		struct passwd	*pwd = getpwuid(getuid());
		if (pwd == NULL)
			strcpy (msg, "UNKNOWN");
		else	strcpy (msg, pwd->pw_name);
		strcat(msg, " ");
	}

	/*----------------------------------------------------------------
	The rest of the command line arguments is copied to the request.
	Variable i still points to the first argument after the options.
	If it gets too long we just cut it off by setting i=argc.

        Not any longer ! Since we changed the port/server arguments to empty strings
        we start now with i=1;
          
	----------------------------------------------------------------*/
	for (i=1; i<argc; i++)
	{
		if ((strlen(msg)+strlen(argv[i]))>= (MAX_MSG-1))
		{
			fprintf(stderr, "Query too complex, please simplify!\n");
			fprintf(stderr, "Arguments after '%s' ignored.\n", argv[i]);
			i = argc;
		}
		else	strcat(msg, argv[i]);
		strcat(msg, " ");
	}

	/*----------------------------------------------------------------
	Now we send the request.
	----------------------------------------------------------------*/
	if (write(sock, msg, strlen(msg)) < strlen(msg))
	{
		fprintf(stderr, "Write Error: server %s at port %d\n",
                     servers[this_serverindex].hostname,
                     servers[this_serverindex].port);
		close(sock);
		exit(1);
	}

	/*----------------------------------------------------------------
	We read from the second until we get an EOF (0) or an error (<0).
	----------------------------------------------------------------*/

	while ((i=read(sock, msg, MAX_MSG)) > 0)
	  { 
	    /* if beginning of string contains error_string */ 

           if (strncmp(msg,error_string,strlen(error_string))==0) 
	   {
             /* removing error_string from msg */ 
             for (k=0;k<(strlen(msg)-strlen(error_string));k++) 
             {
               msg[k]=msg[k+strlen(error_string)];
             }
             for (l=k;k<strlen(msg);l++) 
             { 
                msg[k]='\0'; 
             }
	     fprintf(stderr,msg);
             return_code=1;
           }
           else
           {          
	     if(write(1, msg, i) < 0) {
                fprintf(stderr,"cannot print result to STDOUT\n");
             }
	   }
	}
	if (i<0)
	{
		fprintf(stderr, "Read Error: server %s at port %d\n",
                     servers[this_serverindex].hostname,
                     servers[this_serverindex].port);
		exit(1);
	}

	i=close(sock);

	if (i<0)
	{
		fprintf(stderr, "Close Error: server %s at port %d\n",
                     servers[this_serverindex].hostname,
                     servers[this_serverindex].port);
		exit(1);
	}

	exit(return_code);        /* returning return code which is 1 if error_string was received */
}

static	void usage ( void )
{
	fprintf(stderr, "Usage: phone [-server <hostname>] [-port <portnumber>] [-verbose] <query>\n");
	return;
}


/*--------------------------------------------------------------------
					SIGNAL HANDLING

Upon any signal we suck the output off the stream and terminate.
--------------------------------------------------------------------*/
static  void     CatchSignal(int sgn)
{
/*
	The following piece of code is a bit strange: Bernd's server
	has troubles with connections that break. Therefore, we
	planned to catch the CTRL-C, drain the remaining output and
	terminate gracefully. As it turned out this is very annoying
	for the users. Therefore we decided not to do it.
                
	In case someone wants to implement it later please note that
	the code below is not all you have to do. Between the opening
	the socket and the end of writing the request one musrt not
	wait for draining the socket - this is a deadlock between
	server and client.
*/
/*
	fprintf(stderr, "Interrupt. Please wait for graceful shutdown.\n");
	if (sock >= 0)
	{
		char	msg[MAX_MSG+1];
		while (read(sock, msg, MAX_MSG) > 0);
		close(sock);
		sock = -1;
	}
*/
	fprintf(stderr, "\nUser interrupt.\n");
	exit(0);
}

static	void	SetupSignals()
{
        signal (SIGHUP,  CatchSignal);
        signal (SIGINT,  CatchSignal);
        signal (SIGQUIT, CatchSignal);
        signal (SIGTERM, CatchSignal);
        signal (SIGABRT, CatchSignal);
}


